/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jpeg_io.h"
#include "checksum.h"
#include "common.h"

void jpeg_error_exit (j_common_ptr cinfo)
{
    my_error_ptr myerr = (my_error_ptr) cinfo->err;
    (*cinfo->err->output_message) (cinfo);
    longjmp(myerr->setjmp_buffer, 1);
}

int jpeg_getSize(const char *filename,int *width, int *height, int *grayscale)
{
    struct jpeg_decompress_struct cinfo;
    struct my_error_mgr jerr;
    FILE *file = fopen( filename , "rb");
    if (file == NULL)
        return 0;

    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = jpeg_error_exit;

    if (setjmp(jerr.setjmp_buffer)) {

        jpeg_destroy_decompress(&cinfo);
        fclose(file);
        return 0;
    }

    jpeg_create_decompress(&cinfo);
    jpeg_stdio_src(&cinfo, file);
    jpeg_read_header(&cinfo, TRUE);
    *grayscale=0;
    if (cinfo.jpeg_color_space==JCS_GRAYSCALE)
        *grayscale=1;
    *width=cinfo.image_width;
    *height=cinfo.image_height;
    jpeg_destroy_decompress(&cinfo);
    fclose(file);
    return 1;
}
// Load jpeg file from file
//
FILE_IMAGE *jpeg_load ( const char *filename )
{
    FILE_IMAGE *image=NULL;
    struct jpeg_decompress_struct cinfo;
    struct my_error_mgr jerr;
    int row_stride;
    JSAMPARRAY buffer;
    int c=0;
    
    FILE *file = fopen( filename , "rb");
    if (file == NULL)
        return NULL;
    
    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = jpeg_error_exit;
    //jerr.pub.output_message = jpeg_output_message;
    if (setjmp(jerr.setjmp_buffer)) {
        /* If we get here, the JPEG code has signaled an error.
         * We need to clean up the JPEG object, close the input file, and return.
         */
        jpeg_destroy_decompress(&cinfo);
        fclose(file);
        return NULL;
    }
    jpeg_create_decompress(&cinfo);
    jpeg_stdio_src(&cinfo, file);
    jpeg_read_header(&cinfo, TRUE);
    
    jpeg_start_decompress(&cinfo);
    row_stride = cinfo.output_width * cinfo.output_components;

    image = (FILE_IMAGE *) malloc (sizeof(FILE_IMAGE));
    if (image == NULL)
    {
        jpeg_destroy_decompress(&cinfo);
        fclose(file);
        return NULL;
    }
    image->data = (unsigned char *) malloc ( cinfo.output_width * cinfo.output_height * cinfo.output_components);
    if (image->data == NULL)
    {
        free(image);
        jpeg_destroy_decompress(&cinfo);
        fclose(file);
        return NULL;
    }
    buffer = (*cinfo.mem->alloc_sarray)
    ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
    
   
    while (cinfo.output_scanline < cinfo.output_height) {
        /* jpeg_read_scanlines expects an array of pointers to scanlines.
         * Here the array is only one element long, but you could ask for
         * more than one scanline at a time if that's more convenient.
         */
        (void) jpeg_read_scanlines(&cinfo, buffer, 1);
        /* Assume put_scanline_someplace wants a pointer and sample count. */
        memcpy(image->data+(row_stride*c),buffer[0],row_stride);
        c++;
    }
    image->width = cinfo.output_width;
    image->height = cinfo.output_height;
    image->type = IMAGE_TYPE_JPEG;
    image->bpp=24;
    if (cinfo.jpeg_color_space==JCS_GRAYSCALE)
        image->grayscale=1;
    image->data_size=image->width*image->height*3;
    
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    fclose(file);
    
    return image;
}

FILE_IMAGE *jpeg_load_mem ( MEMORY_FILE_IMAGE *mem_image)
{
    FILE *fj=fopen("tmpimg.bin","wb");
    fwrite(mem_image->data,1,mem_image->data_size,fj);
    fclose(fj);
    FILE_IMAGE*ret_image=jpeg_load("tmpimg.bin");
    unlink("tmpimg.bin");
    return ret_image;
}

FILE_IMAGE *jpeg_load_mem_p ( unsigned char *data, size_t data_size)
{
    MEMORY_FILE_IMAGE image;
    image.data = data;
    image.data_size=data_size;
    return jpeg_load_mem(&image);
}

// Save jpeg file on disk
//
int jpeg_write(FILE_IMAGE *image, const char *filename, int quality,int subsample,int dct_opt,int huffman_opt, int lossless)
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
	FILE *file;
    int row_stride;
    JSAMPLE *scanline;

    if (image->type != IMAGE_TYPE_JPEG)
        return 0;


    cinfo.err = jpeg_std_error(&jerr);
    cinfo.err->trace_level = 0;
    jpeg_create_compress(&cinfo);
    
    file = fopen( filename, "wb");
    if (file == NULL)
    {
        jpeg_destroy_compress(&cinfo);
        return 0;
    }
    jpeg_stdio_dest(&cinfo, file);
    
    cinfo.image_width = image->width; 	/* image width and height, in pixels */
    cinfo.image_height = image->height;
    if (image->grayscale==0)
    {
        cinfo.input_components = 3;		/* # of color components per pixel */
        cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
    }
    else
    {
        cinfo.input_components = 1;		/* # of color components per pixel */
        cinfo.in_color_space = JCS_GRAYSCALE; 	/* colorspace of input image */
    }
    jpeg_set_defaults(&cinfo);

    jpeg_set_quality(&cinfo, quality, 1 );
    if (subsample!=JPEG_SUBSAMPLE_DEFAULT)
    {
        if (subsample==JPEG_SUBSAMPLE_422)
        {
            cinfo.comp_info[0].h_samp_factor = 2;
            cinfo.comp_info[0].v_samp_factor = 1;
            cinfo.comp_info[1].h_samp_factor = 1;
            cinfo.comp_info[1].v_samp_factor = 1;
            cinfo.comp_info[2].h_samp_factor = 1;
            cinfo.comp_info[2].v_samp_factor = 1;
        }
        else if (subsample==JPEG_SUBSAMPLE_411)
        {
            cinfo.comp_info[0].h_samp_factor = 2;
            cinfo.comp_info[0].v_samp_factor = 2;
            cinfo.comp_info[1].h_samp_factor = 1;
            cinfo.comp_info[1].v_samp_factor = 1;
            cinfo.comp_info[2].h_samp_factor = 1;
            cinfo.comp_info[2].v_samp_factor = 1;
        }
        else if (subsample==JPEG_SUBSAMPLE_444)
        {
            cinfo.comp_info[0].h_samp_factor = 1;
            cinfo.comp_info[0].v_samp_factor = 1;
            cinfo.comp_info[1].h_samp_factor = 1;
            cinfo.comp_info[1].v_samp_factor = 1;
            cinfo.comp_info[2].h_samp_factor = 1;
            cinfo.comp_info[2].v_samp_factor = 1;
        }
    }
    cinfo.optimize_coding= huffman_opt; // huffman optimization
    cinfo.dct_method          = dct_opt;
    if (lossless==1)
    {
        // Lossless options

    }

    jpeg_start_compress(&cinfo, 1);
    
    row_stride = image->width * cinfo.input_components;

    scanline=(JSAMPLE*)image->data;
    while (cinfo.next_scanline < cinfo.image_height) {
        (void) jpeg_write_scanlines(&cinfo, &scanline, 1);
        scanline += row_stride;
    }
    jpeg_finish_compress(&cinfo);
    fclose(file);
    jpeg_destroy_compress(&cinfo);
    return 1;
}

void init_destination (j_compress_ptr cinfo) {
	xdpjpg_dst_ptr dest = (xdpjpg_dst_ptr) cinfo->dest;
    
    dest->buffer = (JOCTET *) malloc(OUTPUT_BUF_SIZE *sizeof(JOCTET));
    
	dest->pub.next_output_byte = dest->buffer;
	dest->pub.free_in_buffer = OUTPUT_BUF_SIZE;
    dest->out_buffer_size=0;
}

int empty_output_buffer (j_compress_ptr cinfo) {
	xdpjpg_dst_ptr dest = (xdpjpg_dst_ptr) cinfo->dest;
    memcpy(dest->out_buffer+dest->out_buffer_size,dest->buffer,OUTPUT_BUF_SIZE);
    dest->out_buffer_size+=OUTPUT_BUF_SIZE;
	dest->pub.next_output_byte = dest->buffer;
	dest->pub.free_in_buffer = OUTPUT_BUF_SIZE;
    
	return TRUE;
}

void term_destination (j_compress_ptr cinfo) {
	xdpjpg_dst_ptr dest = (xdpjpg_dst_ptr) cinfo->dest;
    
    size_t datacount = OUTPUT_BUF_SIZE - dest->pub.free_in_buffer;
    
	// write any data remaining in the buffer
    
	if (datacount > 0) {
        memcpy(dest->out_buffer+dest->out_buffer_size,dest->buffer,datacount);
        dest->out_buffer_size+=datacount;
	}
}

void jpeg_xdepth_dst (j_compress_ptr cinfo,  JOCTET *buffer) {
    xdpjpg_dst_ptr dest;

    if (cinfo->dest == NULL) {
        cinfo->dest = (struct jpeg_destination_mgr *)malloc(sizeof(DestinationManager));
    }

    dest = (xdpjpg_dst_ptr) cinfo->dest;
    dest->pub.init_destination = init_destination;
    dest->pub.empty_output_buffer = empty_output_buffer;
    dest->pub.term_destination = term_destination;
    dest->outfile = NULL;
    dest->out_buffer = buffer;
    dest->out_buffer_size=0;
}
// Save jpeg file in memory
//
MEMORY_FILE_IMAGE *jpeg_write_mem(FILE_IMAGE *image, int quality, int subsample, int dct_opt, int huffman_opt, int lossless)
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    int row_stride;
    size_t orig_image_size;
    unsigned char *image_buffer=NULL;
    MEMORY_FILE_IMAGE *retImage=NULL;
    xdpjpg_dst_ptr dest;
	JSAMPLE *scanline;
    if (image->type != IMAGE_TYPE_JPEG)
        return NULL;
    
    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);

    orig_image_size=image->width*image->height*3;
    image_buffer = (unsigned char *) malloc(orig_image_size);

    jpeg_xdepth_dst (&cinfo,  (JOCTET *)image_buffer);

    retImage = (MEMORY_FILE_IMAGE *)malloc(sizeof(MEMORY_FILE_IMAGE));
    if (retImage == NULL)
    {
        jpeg_destroy_compress(&cinfo);
        return NULL;
    }

    memset(retImage,0,sizeof(MEMORY_FILE_IMAGE));
    retImage->width=image->width;
    retImage->height=image->height;

    cinfo.image_width = image->width; 	/* image width and height, in pixels */
    cinfo.image_height = image->height;
    if (image->grayscale==0)
    {
        cinfo.input_components = 3;		/* # of color components per pixel */
        cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
    }
    else
    {
        cinfo.input_components = 1;		/* # of color components per pixel */
        cinfo.in_color_space = JCS_GRAYSCALE; 	/* colorspace of input image */
    }
    if (lossless==1)
    {
        // Lossless options

    }
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, quality, 1 );

    if (subsample!=JPEG_SUBSAMPLE_DEFAULT)
    {
        if (subsample==JPEG_SUBSAMPLE_422)
        {
            cinfo.comp_info[0].h_samp_factor = 2;
            cinfo.comp_info[0].v_samp_factor = 1;
            cinfo.comp_info[1].h_samp_factor = 1;
            cinfo.comp_info[1].v_samp_factor = 1;
            cinfo.comp_info[2].h_samp_factor = 1;
            cinfo.comp_info[2].v_samp_factor = 1;
        }
        else if (subsample==JPEG_SUBSAMPLE_411)
        {
            cinfo.comp_info[0].h_samp_factor = 2;
            cinfo.comp_info[0].v_samp_factor = 2;
            cinfo.comp_info[1].h_samp_factor = 1;
            cinfo.comp_info[1].v_samp_factor = 1;
            cinfo.comp_info[2].h_samp_factor = 1;
            cinfo.comp_info[2].v_samp_factor = 1;
        }
        else if (subsample==JPEG_SUBSAMPLE_444)
        {
            cinfo.comp_info[0].h_samp_factor = 1;
            cinfo.comp_info[0].v_samp_factor = 1;
            cinfo.comp_info[1].h_samp_factor = 1;
            cinfo.comp_info[1].v_samp_factor = 1;
            cinfo.comp_info[2].h_samp_factor = 1;
            cinfo.comp_info[2].v_samp_factor = 1;
        }
    }
    cinfo.optimize_coding = huffman_opt; // huffman optimization
    cinfo.dct_method          = dct_opt;


    jpeg_start_compress(&cinfo, 1);

    row_stride = image->width * cinfo.input_components;

    scanline=(JSAMPLE*)image->data;
    while (cinfo.next_scanline < cinfo.image_height) {
        (void) jpeg_write_scanlines(&cinfo, &scanline, 1);
        scanline += row_stride;
    }

    jpeg_finish_compress(&cinfo);
    dest=(xdpjpg_dst_ptr)cinfo.dest;
    retImage->data=image_buffer;
    retImage->data_size = dest->out_buffer_size;
    jpeg_destroy_compress(&cinfo);
    retImage->type=IMAGE_TYPE_JPEG;
    retImage->bpp=24;
    return retImage;
}

unsigned short computeChecksum(MEMORY_FILE_IMAGE *jpeg_image)
{
    unsigned char b_marker[2];
    unsigned char b_marker_len[2];
    unsigned char *buffer=NULL;
    unsigned short oc0=0;
    unsigned short oc1=0;
    unsigned short c0=0;
    unsigned short c1=0;
    unsigned char *jpegfile=jpeg_image->data;
	unsigned int dtsize;

    unsigned int off=0;
	unsigned short cheksum;
    if (jpegfile[0]==0xFF && jpegfile[1]==0xD8)
    {
        jpegfile+=2;
        off+=2;
        do {
            unsigned short *mark_len;
            b_marker[0]=jpegfile[0];
            b_marker[1]=jpegfile[1];
            if (b_marker[0]!=0xFF)
                break;
            b_marker_len[0]=jpegfile[3];
            b_marker_len[1]=jpegfile[2];

            mark_len=(unsigned short *)b_marker_len;
            if (buffer!=NULL)
                free(buffer);
            buffer=(unsigned char *)malloc(*mark_len);
            if (buffer==NULL)
                return 0;
            memcpy(buffer,jpegfile+4,*mark_len-2);
            if (b_marker[1]==0xDA)
            {
                jpegfile+=*mark_len+2;
                off+=*mark_len+2;
                break;
            }

            jpegfile+=*mark_len+2;
            off+=*mark_len+2;
        }while (b_marker[0]==0xFF);
        dtsize=jpeg_image->data_size-off;
        if (jpeg_image->data[jpeg_image->data_size-2]==0xFF && jpeg_image->data[jpeg_image->data_size-1]==0xD9)
            dtsize-=2;
        oc0=0;
        oc1=0;
        cheksum=Fletcher16(jpegfile,dtsize,c0,c1,&oc0,&oc1);
        return cheksum;
    }

    return 0;
}
