/*
 Copyright (c) 2006-2016 Trellis Management Co. Ltd.

 This software module is an implementation of one or more tools as proposed
 for the JPEG XT standard.
 
 The copyright in this software is being made available under the
 license included with the sources LICENSE.TXT. This software may be subject to other third
 party and contributor rights, including patent rights, and no such
 rights are granted under this license. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "pfm_io.h"

int pfm_get_line(FILE *io,char *buffer, int length) {
	int i;
	memset(buffer, 0, length);
	for(i = 0; i < length; i++) {
        fread(&buffer[i],1,1,io);
		if(buffer[i] == 0x0A)
			break;
	}
	return (i < length) ? 1 : 0;
}

int pfm_get_int(FILE *io) {
    char c = 0;
	int i = 0;
    int firstchar;
    
    // skip forward to start of next number
    fread(&c,1,1,io);
    while (1) {
        // eat comments
        
        if (c == '#') {
			// if we're at a comment, read to end of line
            
            firstchar = 1;
            
            while (1) {
                fread(&c,1,1,io);

                if (firstchar==1 && c == ' ') {
					// loop off 1 sp after #
                    
                    firstchar = 0;
				} else if (c == '\n') {
					break;
				}
			}
		}
        
        if (c >= '0' && c <='9') {
			// we've found what we were looking for
            
            break;
		}
        fread(&c,1,1,io);
    }
    
    // we're at the start of a number, continue until we hit a non-number
    
    
    
    while (1) {
        i = (i * 10) + (c - '0');
        
        fread(&c,1,1,io);
        
        if (c < '0' || c > '9')
            break;
    }
    
    return i;
}

int pfm_channels(const char *filename)
{
    char fileType[3];
    FILE *file = fopen(filename,"rb");
    if (file == NULL)
        return 0;
    fread(fileType,1,2,file);
    fileType[2]=0;
    fclose(file);
    if (!strcmp(fileType,"PF"))
        return 3;
    else
        return 1;
}

FILE_IMAGE *pfm_load(const char *filename)
{
    char fileType[3];
    char line_buffer[PFM_MAXLINE];
    int imageType=0;
    unsigned width,height;
    float scalefactor=1;
    int result;
    unsigned lineWidth;
    float *lineBuffer = NULL;
    unsigned y,x;
    FILE_IMAGE *image=NULL;
    unsigned input_channels=3;
    FILE *file = fopen(filename,"rb");
    if (file == NULL)
        return NULL;
    fread(fileType,1,2,file);
    fileType[2]=0;
    if (!strcmp(fileType,"PF"))
        imageType=PFM_TYPE_RGBF;
    else
    {
        imageType=PFM_TYPE_FLOAT;
        input_channels=1;
        //fclose(file);
        //return NULL;
    }
    width = pfm_get_int(file);
    height = pfm_get_int(file);
    
    result=pfm_get_line(file,line_buffer,PFM_MAXLINE);
    if (result==1)
    {
        result=(sscanf(line_buffer, "%f", &scalefactor) == 1) ? 1 : 0;
    }
    if (result==0)
    {
        fclose(file);
        return NULL;
    }
    lineWidth = input_channels*width;


    image=(FILE_IMAGE *) malloc (sizeof(FILE_IMAGE));
    if (image == NULL)
    {
        fclose(file);
        return NULL;
    }
    image->data = (unsigned char *) malloc( (width * height * input_channels) * sizeof(float));
    if (image->data == NULL)
    {
        fclose(file);
        free (image);
        return NULL;
    }
    image->width = width;
    image->height = height;
    image->type = IMAGE_TYPE_HDR;
    image->bpp=96;
    if (input_channels==1)
        image->grayscale=1;
    
    lineBuffer = (float*)malloc(lineWidth * sizeof(float));
    if(lineBuffer == NULL) {
        fclose(file);
        return NULL;
    }
    if (input_channels==3)
    {
        for (y = 0; y < height; y++) {
            float *channel;
            unsigned pitch;
            IMAGE_PIXELF *pScanline;
            fread(lineBuffer,sizeof(float),lineWidth,file);
            channel = lineBuffer;
            pitch=lineWidth*sizeof(float);
            pScanline = (IMAGE_PIXELF *)get_scanline(image,y);
            if(scalefactor > 0) {
                for ( x = 0; x < width; x++) {
                    REVERSEBYTES(channel++, &pScanline[x].red);
                    REVERSEBYTES(channel++, &pScanline[x].green);
                    REVERSEBYTES(channel++, &pScanline[x].blue);
                }
            }
            else {
                for ( x = 0; x < width; x++) {
                    pScanline[x].red		= *channel++;
                    pScanline[x].green	= *channel++;
                    pScanline[x].blue	= *channel++;
                }

            }
        }
    }
    else
    {
        for (y = 0; y < height; y++) {
            float *channel;
            unsigned pitch;
            IMAGE_PIXELF_BW *pScanline;
            fread(lineBuffer,sizeof(float),lineWidth,file);
            channel = lineBuffer;
            pitch=lineWidth*sizeof(float);
            pScanline = (IMAGE_PIXELF_BW *)get_scanline(image,y);
            if(scalefactor > 0) {
                for ( x = 0; x < width; x++) {
                    REVERSEBYTES(channel++, &pScanline[x].gray);

                }
            }
            else {
                for ( x = 0; x < width; x++) {
                    pScanline[x].gray		= *channel++;

                }

            }
        }

    }
    image->data_size=(image->width*image->height*input_channels)*sizeof(float);
     
    free(lineBuffer);
    fclose(file);
     
    return image;
}

int pfm_write(FILE_IMAGE *image,const char *filename)
{
    unsigned width,height,lineWidth;
    // save image as Little Endian
    const float scalefactor = -1.0F;
    char buffer[PFM_MAXLINE];
	FILE *file;
    unsigned input_channles=3;
    char magic = 'F';
    unsigned y;
    if (image == NULL)
        return 0;
    //file = fopen(filename,"wb");
    //if (file == NULL)
    //    return 0;
    width=image->width;
    height=image->height;
    if (image->grayscale==1)
    {
        magic='f';
        input_channles=1;
    }
    lineWidth=width*input_channles;

    file=fopen(filename,"wb");
    if (file == NULL)
        return 0;

    sprintf(buffer, "P%c\n%d %d\n%f\n", magic, width, height, scalefactor);
    fwrite(buffer,(unsigned int)strlen(buffer), 1,file);
    fwrite(image->data,1,image->data_size,file);
    fclose(file);
    /*
    // Write the image data
    for (y = 0; y < height; y++) {
            unsigned char *bits = get_scanline(image, y);
            fwrite(bits,1,lineWidth*sizeof(float),file);
        }
    fclose(file);
    */
   return 1;
}
